@@ -6,7 +6,7 @@ Parameterverwaltung
     bedient die meisten Laufzeitparameter
 '''
 import sys
-import getopt
+#import getopt
 import types
 import inspect
 from pathlib import Path, PurePath
@@ -14,8 +14,13 @@ import json
 import types
 import copy
 
-
-
+try:
+    import json5
+    JsonLoads = json5.loads
+    JsonLoad = json5.load
+except:
+    JsonLoads = json.loads
+    JsonLoad = json.load
 
 
 class Param(dict):
@@ -38,12 +43,14 @@ s exception is raised if there is an declaration error within the
 parameters of the class.
         '''
         pass
+
     class ParamError(__ExceptionTemplate):
         '''
 s exception is raised if there is an error within the runtime-parameters.
 s os only within the "Process"-function.
         '''
         pass
+
     class PathEncoder(json.JSONEncoder):
         def default(self, z):
             if isinstance(z, Path) or isinstance(z, PurePath):
@@ -77,6 +84,8 @@ s os only within the "Process"-function.
                 # only parameters with defaults or on the commandline
                 # are in the dictionary
     __DoTranslate = None    # Translation routine for error-messages'
+    __Prefix = None
+    __AllowProcessToExit = True
 
     __WorkPars = {
         'shortpar':     's',
@@ -116,7 +125,9 @@ s os only within the "Process"-function.
             AllParams: bool = True, 
             UserPars:dict = None, 
             UserModes: dict = None,
-            Translate = None):
+            Translate = None,
+            Prefix = None,
+            AllowProcessToExit = True):
         """ The construktor
         Args:
             Def (dict, optional): See SetDef(). Defaults to {}.
@@ -127,9 +138,22 @@ s os only within the "Process"-function.
             AllParams (Boolean, optional): See SetAllParams. Defaults to True.
             UserPars (dict, optional): See SetUserKeys. Defaults to None.
             UserModes (dict, optional): See SetUserKeys. Defaults to None.
+            Prefix (str, optional): When given long params in the form "--xxx.yyyy" are only used
+                    if xxx is equal to the Prefix-value. All other options of this form are silently ignored.
+                    Defaults to None.
+            AllowProcessToExit (bool, optional): If True (the default) the Proces-function exits the program if
+                    there is a "Help" or a "Export" option is set. Otherwise it returns True if these options 
+                    are used and else False.
+                    Defaults to True.
         """
         super(Param, self).__init__()       # Init parent -> make me a dict
         # set the parameters with the individual functions
+        if type(Prefix) == str: 
+            self.__Prefix = Prefix
+        else:
+            self.__Prefix = None
+        self.__AllowProcessToExit = AllowProcessToExit
+
         self.SetDesc(Desc)
         self.SetUserKeys(UserPars = UserPars,UserModes = UserModes)
         self.SetDef(Def)
@@ -449,8 +473,11 @@ s os only within the "Process"-function.
             ShortLen (int): Max. length of the "short"-options (0 or 1)
             LongLen (int): Max. length of the "long"-options
         """
-
-        Text = f"Usage:\n{self.__MyProgName} OPTIONS {self.__AddPar}\n\n{self.__Description}\n\nOptions:\n"
+        if self.__Prefix is None:
+            wPrefText = ''
+        else:
+            wPrefText = f'active prefix = "{self.__Prefix}." only for long (--) options\n\n'
+        Text = f"Usage:\n{self.__MyProgName} OPTIONS {self.__AddPar}\n\n{self.__Description}\n\n{wPrefText}Options:\n"
         for Single in self.__UsageTextList:
             Ut_Short = Single[0]
             Ut_Long = Single[1]
@@ -464,6 +491,8 @@ s os only within the "Process"-function.
             Ut_Text = Single[5].splitlines()
             sl = ShortLen + 3 + 1
             ll = LongLen + 3 + 2
+            if self.__Prefix is not None:
+                ll = ll + len(self.__Prefix) + 3
             Lines = max(len(Ut_Short),len(Ut_Long),len(Ut_Text)+1)
             while len(Ut_Short) < Lines:
                 Ut_Short.append(" ")
@@ -478,6 +507,8 @@ s os only within the "Process"-function.
                 wLine = "\n   "
                 s = Ut_Short[i]
                 l = Ut_Long[i]
+                if self.__Prefix is not None and l != " ":
+                    l = '[' + self.__Prefix + '.]' + l
                 t = Ut_Text[i]
                 if s == " ":
                     n = " " * sl
@@ -752,11 +783,13 @@ s os only within the "Process"-function.
             return f"No action defined for {Param}"
         if Type == "Required":
             return f"{Param} ({OptList}) required but not given"
+        if Type == "Prefix":
+            return f"Error in prefixed parameter {Param}"
         return f"Undefined error Type='{Type}', Param='{Param}', Path='{Path}', FullPath='{FullPath}', Msg='{Msg}', OptList='{OptList}'"
 
 
 
-    def Process(self) -> None:
+    def Process(self, DispName = None, Partial = False) -> bool:
         """
         Process the runtime-arguments
 
@@ -766,18 +799,48 @@ s os only within the "Process"-function.
         """
         if not self.__IsPrepared:
             self.__Prepare()
+        PreList = []
+        for wPar in self.__Argumente[1:]:
+            if wPar[0:2] == '--' and '.' in wPar:
+                xPre = wPar[2:].split('.')[0]
+                if xPre not in PreList:
+                    PreList.append(xPre)
+        wLongList = []
+        for nLong in self.__LongList:
+            wLongList.append(nLong)
+        if len(PreList) > 0:
+            for nPre in PreList:
+                for nLong in self.__LongList:
+                    wLongList.append(nPre + '.' + nLong)
         try:
-            opts, args = getopt.getopt(self.__Argumente[1:],self.__ShortList,self.__LongList)
-        except getopt.GetoptError as exc:
+#            opts, args = getopt.getopt(self.__Argumente[1:],self.__ShortList,self.__LongList)
+            opts, args = getopt(self.__Argumente[1:], self.__ShortList, wLongList, True)
+        except GetoptError as exc:
             wMsg = exc.msg
             raise self.ParamError(wMsg) from None
         self.__RemainArgs = args
         for o, a in opts:
+            if '.' in o:
+                wList = o[2:].split('.')
+                if len(wList) != 2:
+                    raise self.ParamError(self.__MakeErrorMsg(Type="Prefix",Param=o)) from None
+                if wList[0] == self.__Prefix:
+                    o = '--' + wList[1]
             if o in self.__HelpList:
+                if DispName is not None:
+                    print(f"#{'-'*60}\n# {DispName}\n#{'-'*60}\n")
                 print(self.Usage())
-                sys.exit(0)
+                if self.__AllowProcessToExit:
+                    sys.exit(0)
+                return True
 
         for o, a in opts:
+            if '.' in o:
+                wList = o[2:].split('.')
+                if len(wList) != 2:
+                    raise self.ParamError(self.__MakeErrorMsg(Type="Prefix",Param=o)) from None
+                if wList[0] == self.__Prefix:
+                    o = '--' + wList[1]
             if o in self.__ImportList:
                 try:
                     n = Path(a).resolve()
@@ -786,7 +849,8 @@ s os only within the "Process"-function.
                 if n.exists():
                     if n.is_file():
                         try:
-                            wDict = json.load(n.open())
+#                            wDict = json.load(n.open())
+                            wDict = JsonLoad(n.open())
                         except Exception as exc:
                             wMsg = str(exc)
                             raise self.ParamError(self.__MakeErrorMsg(Type="ErrMsg",Msg=wMsg,Path=a,FullPath=n,Param=o)) from None
@@ -801,6 +865,14 @@ s os only within the "Process"-function.
                     raise self.ParamError(self.__MakeErrorMsg(Type="NoPath",Path=a,FullPath=n,Param=o)) from None
 
         for o, a in opts:
+            if '.' in o:
+                wList = o[2:].split('.')
+                if len(wList) != 2:
+                    raise self.ParamError(self.__MakeErrorMsg(Type="Prefix",Param=o)) from None
+                if wList[0] == self.__Prefix:
+                    o = '--' + wList[1]
+                else:
+                    continue
             if o in self.__HelpList:
                 continue
             if o in self.__ImportList:
@@ -810,11 +882,13 @@ s os only within the "Process"-function.
             try:
                 ParName = self.__ParDict[o] 
             except:
-                raise RuntimeError("Internal error, option {o} not found in ParDict")
+                if Partial:
+                    continue
+                raise RuntimeError(f"Error, option {o} not found in ParDict")
             try:
                 wPar = self.__Definition[ParName]
             except:
-                raise RuntimeError("Internal error, option {ParName} not found in Definition")
+                raise RuntimeError(f"Internal error, option {ParName} not found in Definition")
             if self.__WorkPars['needoption'] in wPar.keys():
                 if wPar[self.__WorkPars['needoption']]:
                     Res = self.__CheckOption(ParName,o,wPar,a)
@@ -833,9 +907,19 @@ s os only within the "Process"-function.
                 raise self.ParamError(self.__MakeErrorMsg(Type="NoAct",Param=ParName))
 
         for o, a in opts:
+            if '.' in o:
+                wList = o.split('.')
+                if len(wList) != 2:
+                    raise self.ParamError(self.__MakeErrorMsg(Type="Prefix",Param=o)) from None
+                if wList[0] == self.__Prefix:
+                    o = wList[1]
             if o in self.__ExportList:
+                if DispName is not None:
+                    print(f"//{'-'*60}\n// {DispName}\n//{'-'*60}\n")
                 print(json.dumps(self, sort_keys=True, indent=4, cls=self.PathEncoder))
-                sys.exit(0)
+                if self.__AllowProcessToExit:
+                    sys.exit(0)
+                return True
 
         for i in self.__Definition.keys():
             v = self.__Definition[i]
@@ -845,6 +929,7 @@ s os only within the "Process"-function.
             if Req:
                 if not i in self.keys():
                     raise self.ParamError(self.__MakeErrorMsg(Type="Required",Param=i,OptList=self.__GetOptList(i))) from None
+        return False
 
     def __GetOptList(self,Name: str) -> str:
         """ liste der möglichen Parameter eines Keys"""
@@ -1081,8 +1166,255 @@ s os only within the "Process"-function.
         """
         return self.__ParDict
 
+class MultiParam(dict):
+    class __ExceptionTemplate(Exception):
+        def __call__(self, *args):
+                return self.__class__(*(self.args + args))
+
+        def __str__(self):
+                return ': '.join(self.args)
+
+    class DeclarationError(__ExceptionTemplate):
+        '''
+s exception is raised if there is an declaration error within the 
+parameters of the class.
+        '''
+        pass
+
+    class ParamError(__ExceptionTemplate):
+        '''
+s exception is raised if there is an error within the runtime-parameters.
+s os only within the "Process"-function.
+        '''
+        pass
+
+    def __init__(self, ParDict: dict = None, Global: dict = None):
+        super().__init__()       # Init parent -> make me a dict
+        if type(ParDict) != dict:
+           raise TypeError('ParDict is not a dict') 
+        if Global is not None:
+            if type(Global) != dict:
+                raise TypeError('Global is not a dict') 
+
+        self.__ParDict = ParDict
+        for ParName in self.__ParDict.keys():         
+            ParVal = self.__ParDict[ParName]
+            if type(ParVal) != dict:
+                raise TypeError(f"ParDict[{ParName}] is not a dict") 
+            if 'Def' not in ParVal:
+                raise TypeError(f"ParDict[{ParName}] idoes not include 'Def'")
+            ParDef = ParVal['Def']
+            if type(ParDef) != dict:
+                raise TypeError(f"ParDict[{ParName}]['Def'] is not a dict") 
+
+            if 'Args' not in ParVal:
+                raise TypeError(f"ParDict[{ParName}] idoes not include 'Args'") 
+            ParArgs = copy.deepcopy(ParVal['Args'])
+            if type(ParArgs) != dict:
+                raise TypeError(f"ParDict[{ParName}]['Args'] is not a dict") 
+            ParArgs['AllowProcessToExit'] = False 
+            self[ParName] = Param(Def = ParDef, **ParArgs)
+
+    def Process(self):
+        IsError = False
+        ErrMsg = ''
+        Result = False
+        for ParName in self.keys():
+            SubPar = self[ParName]
+            try:
+                Erg = SubPar.Process(ParName, Partial = True)
+                if Erg:
+                    Result = True
+            except Exception as exc:
+                IsError = True
+                ErrMsg += f"{ParName}: {str(exc)}\n"
+        if IsError:
+            raise self.ParamError(ErrMsg) from None
+        return Result
+                
 
 
+
+class GetoptError(Exception):
+    opt = ''
+    msg = ''
+    def __init__(self, msg, opt=''):
+        self.msg = msg
+        self.opt = opt
+        Exception.__init__(self, msg, opt)
+
+    def __str__(self):
+        return self.msg
+
+#error = GetoptError # backward compatibility
+
+def getopt(args, shortopts, longopts = [], AcceptAll = False):
+    """getopt(args, options[, long_options]) -> opts, args
+
+    Parses command line options and parameter list.  args is the
+    argument list to be parsed, without the leading reference to the
+    running program.  Typically, this means "sys.argv[1:]".  shortopts
+    is the string of option letters that the script wants to
+    recognize, with options that require an argument followed by a
+    colon (i.e., the same format that Unix getopt() uses).  If
+    specified, longopts is a list of strings with the names of the
+    long options which should be supported.  The leading '--'
+    characters should not be included in the option name.  Options
+    which require an argument should be followed by an equal sign
+    ('=').
+
+    The return value consists of two elements: the first is a list of
+    (option, value) pairs; the second is the list of program arguments
+    left after the option list was stripped (this is a trailing slice
+    of the first argument).  Each option-and-value pair returned has
+    the option as its first element, prefixed with a hyphen (e.g.,
+    '-x'), and the option argument as its second element, or an empty
+    string if the option has no argument.  The options occur in the
+    list in the same order in which they were found, thus allowing
+    multiple occurrences.  Long and short options may be mixed.
+
+    """
+
+    opts = []
+    if isinstance(longopts, str):
+        longopts = [longopts]
+    else:
+        longopts = list(longopts)
+    while args and args[0].startswith('-') and args[0] != '-':
+        if args[0] == '--':
+            args = args[1:]
+            break
+        if args[0].startswith('--'):
+            opts, args = do_longs(opts, args[0][2:], longopts, args[1:],AcceptAll)
+        else:
+            opts, args = do_shorts(opts, args[0][1:], shortopts, args[1:],AcceptAll)
+
+    return opts, args
+
+def gnu_getopt(args, shortopts, longopts = [], AcceptAll = False):
+    """getopt(args, options[, long_options]) -> opts, args
+
+    This function works like getopt(), except that GNU style scanning
+    mode is used by default. This means that option and non-option
+    arguments may be intermixed. The getopt() function stops
+    processing options as soon as a non-option argument is
+    encountered.
+
+    If the first character of the option string is `+', or if the
+    environment variable POSIXLY_CORRECT is set, then option
+    processing stops as soon as a non-option argument is encountered.
+
+    """
+
+    opts = []
+    prog_args = []
+    if isinstance(longopts, str):
+        longopts = [longopts]
+    else:
+        longopts = list(longopts)
+
+    # Allow options after non-option arguments?
+    if shortopts.startswith('+'):
+        shortopts = shortopts[1:]
+        all_options_first = True
+    elif os.environ.get("POSIXLY_CORRECT"):
+        all_options_first = True
+    else:
+        all_options_first = False
+
+    while args:
+        if args[0] == '--':
+            prog_args += args[1:]
+            break
+
+        if args[0][:2] == '--':
+            opts, args = do_longs(opts, args[0][2:], longopts, args[1:], AcceptAll)
+        elif args[0][:1] == '-' and args[0] != '-':
+            opts, args = do_shorts(opts, args[0][1:], shortopts, args[1:], AcceptAll)
+        else:
+            if all_options_first:
+                prog_args += args
+                break
+            else:
+                prog_args.append(args[0])
+                args = args[1:]
+
+    return opts, prog_args
+
+def do_longs(opts, opt, longopts, args, AcceptAll = False):
+    try:
+        i = opt.index('=')
+    except ValueError:
+        optarg = None
+    else:
+        opt, optarg = opt[:i], opt[i+1:]
+    if AcceptAll:
+        try:
+            has_arg, opt = long_has_args(opt, longopts)
+        except GetoptError as exc:
+            return opts, args
+    else:
+        has_arg, opt = long_has_args(opt, longopts)
+    if has_arg:
+        if optarg is None:
+            if not args:
+                raise GetoptError('option --%s requires argument' % opt, opt)
+            optarg, args = args[0], args[1:]
+    elif optarg is not None:
+        raise GetoptError('option --%s must not have an argument' % opt, opt)
+    opts.append(('--' + opt, optarg or ''))
+    return opts, args
+
+# Return:
+#   has_arg?
+#   full option name
+def long_has_args(opt, longopts):
+    possibilities = [o for o in longopts if o.startswith(opt)]
+    if not possibilities:
+        raise GetoptError('option --%s not recognized' % opt, opt)
+    # Is there an exact match?
+    if opt in possibilities:
+        return False, opt
+    elif opt + '=' in possibilities:
+        return True, opt
+    # No exact match, so better be unique.
+    if len(possibilities) > 1:
+        # XXX since possibilities contains all valid continuations, might be
+        # nice to work them into the error msg
+        raise GetoptError('option --%s not a unique prefix' % opt, opt)
+    assert len(possibilities) == 1
+    unique_match = possibilities[0]
+    has_arg = unique_match.endswith('=')
+    if has_arg:
+        unique_match = unique_match[:-1]
+    return has_arg, unique_match
+
+def do_shorts(opts, optstring, shortopts, args, AcceptAll = False):
+    while optstring != '':
+        opt, optstring = optstring[0], optstring[1:]
+        if AcceptAll:
+            try:
+                wHasArgs = short_has_arg(opt, shortopts)
+            except GetoptError as exc:
+                wHasArgs = False
+        else:
+            wHasArgs = short_has_arg(opt, shortopts)
+        if wHasArgs:
+            if optstring == '':
+                if not args:
+                    raise GetoptError('option -%s requires argument' % opt, opt)
+                optstring, args = args[0], args[1:]
+            optarg, optstring = optstring, ''
+        else:
+            optarg = ''
+        opts.append(('-' + opt, optarg))
+    return opts, args
+
+def short_has_arg(opt, shortopts):
+    for i in range(len(shortopts)):
+        if opt == shortopts[i] != ':':
+            return shortopts.startswith(':', i+1)
+    raise GetoptError('option -%s not recognized' % opt, opt)
     
 if __name__ == "__main__":
 
@@ -1156,7 +1488,30 @@ if __name__ == "__main__":
                 'm': 'C',
                 'd': 'Mehrmals zum hochzählen'},
         }
-
+    TestDef_2 =     {
+        'Help': {   
+                's': 'h',
+                'l': 'help',
+                'm': 'H',
+                'd': 'Diesen Hilfetext anzeigen und beenden'},
+        'Export': { 
+                's': 'x',
+                'l': 'export',
+                'm': 'X',
+                'd': 'Ausgabe der aktuellen Konfiguration und Beenden'},
+        'ConfFile': {
+                's': 'z',
+                'l': 'par',
+                'm': 'x',
+                'd': '''Zuerst die Werte aus der Datei lesen, 
+    danach erst die Komandozeilenparameter'''},
+        'Verbose': {
+                's': 'v',
+                'l': 'verbose',
+                'r': False,
+                'm': 'b',
+                'd': 'Sei gesprächig'},
+        }
 
     import shlex
 
@@ -1177,11 +1532,53 @@ if __name__ == "__main__":
 
 
     def main():
+
+        m = MultiParam({ 
+            'Alpha': 
+                { 
+                'Def': TestDef_1, 
+                'Args':
+                    { 
+                    'Desc': "Dies ist ein Test\ndas bedeutet hier steht nur\nnonsens", 
+                    'AddPar': "File .... File", 
+                    'Translate': Trans,
+                    'AllParams': True,
+                    'Prefix': 'alpha'
+                    }
+                },
+            'Beta':
+                {
+                'Def': TestDef_2, 
+                'Args':
+                    {
+                    'Desc': "Dies ist ein Test # 2", 
+                    'AllParams': True,
+                    'Prefix': 'beta'
+                    }
+                }
+            })
+        try:
+            Erg = m.Process()
+        except Exception as exc:
+            dir(exc)
+            print(exc)
+            return
+        if Erg:
+            return
+        for Name, Par in m.items():
+            print(f"{'-' * 60}\n{Name}\n{'-' * 60}")
+            for key,value in Par.items():
+                print(f"{Name} -> {key}: {value}")
+
+
+
+
         a = Param(Def = TestDef_1, 
             Desc = "Dies ist ein Test\ndas bedeutet hier steht nur\nnonsens", 
             AddPar = "File .... File", 
             Translate = Trans,
             AllParams = True,
+            AllowProcessToExit = False
             )
         # a.SetArgs(Args = shlex.split('Test -v -CCC -f /Mist --dir=/tmp'))
         try:
@@ -1197,5 +1594,29 @@ if __name__ == "__main__":
         print(f"Rest: {a.GetRemainder()}")
         print(a.GetCmdPar('Text'))
         print(a.Definition)
-        # print(a.Usage())
+        print(a.Usage())
+        print('-'*60)
+        b = Param(Def = TestDef_2, 
+            Desc = "Dies ist ein Test # 2", 
+#            AddPar = "File .... File", 
+            Translate = Trans,
+            AllParams = True,
+            Prefix = 'pre'
+            )
+#        b.SetArgs(Args = shlex.split('Test --ignore.help -v -x'))
+        try:
+            b.Process()
+        except Exception as exc:
+            dir(exc)
+            print(exc)
+            return
+        for key,value in b.items():
+            print(key,value)
+        # print(dir(a))
+        print(f"Ergebnis: {b}")
+        print(f"Rest: {b.GetRemainder()}")
+        print(b.GetCmdPar('Text'))
+        print(b.Definition)
+        print(b.Usage())
+        print('-'*60)
     main()
